/*-
 * Copyright (C)2013 @BABOLO http://www.babolo.ru/
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C)2013 @BABOLO http://www.babolo.ru/\n\
@(#)All rights reserved.\n\
@(#)V.M "VERS;
static const char rcsid[] = "$Id: rcvfile2045.c,v 1.3 2014/10/07 13:09:47 babolo Exp $";
#endif /* not lint */

#define BLIN_COMPAT  2
#define MIFE_COMPAT  3
#define Bpars_COMPAT 2

#include <sys/types.h>
#include <sys/event.h>
#include <sys/time.h>
#include <sysexits.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <fcntl.h>
#include <stdio.h>
#include <errno.h>
#include <err.h>
#include <babolo/BLINflag.h>
#include <babolo/parser.h>
#include <mife.h>

typedef struct {
    BLIN_flag   flags;
#       define RCVFILE2045_RN 0x800000 /*   "\r\n"                               */
#       define RCVFILE2045_DU 0x400000 /*        */
#       define RCVFILE2045_DI 0x400000 /*          */
    babolo_opt *opt;
    char       *fford;
} rcvfile2045_cf;

static void
usage(int ex) {
    fprintf( stderr
           , "rcvfile2045 @BABOLO V.M "VERS"  "DATE" http://www.babolo.ru/\n"
             "    -q        - quiet\n"
             "    -v        - verbose\n"
           )
    ;
    exit(ex);
}


static char *
rediropt(rcvfile2045_cf *options, int redirenv) {
    ifBLIN_FLAG = options->flags;
    char *e, *o;

    if  (!(o = babolo_getoptarg(&options->opt, 0))) {
        ifBLIN_QW0("No argument");
        usage(EX_USAGE);
    } else if ((redirenv > 0) && (!(o = getenv(e = o)))) {
        ifBLIN_QX0("No ENV %s", e);
        usage(EX_USAGE);
    }
    return(o);
}

int
main(int argc, char *argv[], char **envp) {
#   define blin_internal_flags (options->flags)
    int c, redirenv, fd, ex = EX_OK;
    char *flar, *o, *inchars, *q, *h, *inend;
    rcvfile2045_cf *options;
    mife_descriptor *md;
    babolo_parm *bp;
    ssize_t alll;

    if  (!(options = calloc(1, sizeof(*options)))) {
        warnx("Malloc failed #1");
        errno = ENOMEM;
        ex = EX_OSERR;
        goto out;
    }
    options->flags = BLIN_VER1;
    if  (babolo_openopt(Bpars_NONU, &options->opt, argc, argv, flar = "dD:eEqvz:Z:?")) {
        ifBLIN_QW0("openopt");
        ex = EX_SOFTWARE;
        goto out;
    }
    for (redirenv = 0; (c = babolo_getopt(&options->opt)) != 0; redirenv--) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'd': options->flags |= RCVFILE2045_DU;                                              break;
        case 'D': options->fford = rediropt(options, redirenv);                                  break;
        case 'e': options->flags |= RCVFILE2045_DI;                                              break;
        case 'E': redirenv = 2;                                                                  break;
        case 'q': options->flags &= ~BLIN_VERMASK;                                               break;
        case 'v': BLIN_VERBOSE(options->flags);                                                  break;
        case 'z': o = rediropt(options, redirenv);
                ; bp = babolo_getparm(Bpars_NOEN | Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &options->opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'Z': o = rediropt(options, redirenv);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &options->opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case '?': usage(EX_OK);
        default : ifBLIN_QX0("Illegal flag -%c", c);
                ; usage(EX_USAGE);
    }   }
    if  (!(md = mife_opef(0, fileno(stdin)))) {
        ifBLIN_QW0("mife_opef");
        ex = EX_IOERR;
        goto out;
    }
    if  (0 > (alll = mife_read(md, 0, 0))) {
        ifBLIN_QW0("mife_read");
        ex = EX_IOERR;
        goto out;
    }
    if  (!(inchars = mife_get(md, 0))) {
        ifBLIN_QW0("mife_get");
        ex = EX_IOERR;
        goto out;
    }
    inend = inchars + alll;
    if  (options->fford) {
        if  (0 > (fd = open(options->fford, O_WRONLY | O_CREAT | O_TRUNC, 0666))) {
            ifBLIN_QW0("open");
            ex = EX_IOERR;
            goto out;
        }
        if  (0 > mife_writ(fd, inchars, alll)) {
            ifBLIN_QX0("mife_writ d");
            ex = EX_IOERR;
            goto out;
        }
        close(fd);
    }
    if  (!(o = index(inchars, '\n')) || (o == inchars)) {
        o = inchars + ((alll < 16) ? alll : 16);
        *o = '\0';
        ifBLIN_QX0("Input string?=%s~", inchars);
        ex = EX_DATAERR;
        goto out;
    }
    /* o        */
    if  (*--o == '\r') {
        options->flags |= RCVFILE2045_RN;
        *(o++) = '\0';
    } else {
        ++o;
    }
    *o = '\0';
    /*      */
    if  (strncmp(inchars, "--", 2)) {
        ifBLIN_QX0("Input mark?=%s~", inchars);
        ex = EX_DATAERR;
        goto out;
    }
    while (o < inend - 2) { /* XXXX if ... else ERROR */
        ++o;
        if  (!(options->flags & RCVFILE2045_RN) && !strncmp(o, "\n", 1)) {
            ++o;
            break;
        }
        if  ((options->flags & RCVFILE2045_RN) && !strncmp(o, "\r\n", 2)) {
            o += 2;
            break;
        }
        if  (!(o = index(o, '\n'))) {
            ifBLIN_QX0("No input file begin");
            ex = EX_DATAERR;
            goto out;
    }   }
    /* o       */
    if  (0 > asprintf(&h, (options->flags & RCVFILE2045_RN) ? "\r\n%s--" : "\n%s--", inchars)) {
        ifBLIN_QX0("asprintf");
        ex = EX_IOERR;
        goto out;
    }
    ifBLIN_QX2("END=%s~\n", h);
    if  (!(q = memmem(o, inend - o, h, strlen(h)))) {
        int i;

        ifBLIN_QX0("No input file end [%u>%u]", inend - inchars, inend - o);
        if  (options->flags & RCVFILE2045_DI) {
            i = inend - o - strlen(h);
        } else {
            i = 0;
        }
        if  (options->flags & RCVFILE2045_DU) {
            for (; i < inend - o; ++i) {
                if  (!(i & 0xF)) fprintf(stderr, "%06X:", i);
                if  (!(i & 0x1)) fprintf(stderr, " ");
                if  (!(i & 0x3)) fprintf(stderr, " ");
                fprintf(stderr, "%02X", o[i] & 0xFF);
                if  ((i & 0xF) == 0xF) fprintf(stderr, "\n");
        }   }
        ex = EX_DATAERR;
        goto out;
    }
    if  (0 > mife_writ(fileno(stdout), o, q - o)) {
        ifBLIN_QX0("mife_writ");
        ex = EX_IOERR;
        goto out;
    }
out:
    return(ex);
}
